"""Copyright © 2023 Pixelgen Technologies AB."""

from __future__ import annotations

import typing
from pathlib import Path

from . import (
    BaseAdapterQCTestsMixin,
    BaseAmpliconTestsMixin,
    BaseAnalysisTestsMixin,
    BaseAnnotateTestsMixin,
    BaseCollapseTestsMixin,
    BaseDemuxTestsMixin,
    BaseGraphTestsMixin,
    BaseLayoutTestsMixin,
    BasePreQCTestsMixin,
    BaseReportTestsMixin,
)

if typing.TYPE_CHECKING:
    from .workflow_context import PixelatorWorkflowContext


# IMPORTANT: Note the order of the base classes.
# This is important to ensure that dependencies are resolved correctly.
# A mixin with dependencies on test cases defined in other mixins must be
# listed **before** the dependent mixins.
class PixelatorWorkflowTest(
    BaseReportTestsMixin,
    BaseAnalysisTestsMixin,
    BaseLayoutTestsMixin,
    BaseAnnotateTestsMixin,
    BaseGraphTestsMixin,
    BaseCollapseTestsMixin,
    BaseDemuxTestsMixin,
    BaseAdapterQCTestsMixin,
    BasePreQCTestsMixin,
    BaseAmpliconTestsMixin,
):
    """Pixelator workflow test base class.

    This class bundles all the basic tests for each command in the pixelator workflow.
    It is intended to be used as a base class for new workflow tests.
    New tests cases per command can be added by creating a new method and marking it
    with a dependency decorator. The command output of each stage is called
    test_<stage name>_run.

    So for a test case testing annotate output:

    @pytest.mark.dependency(depends=["test_annotate_run"])
    def test_some_annotate_output(self):
        pass

    The workflow test is configurable by setting class variables in the subclass.
    The current mixin plugins are defined to use the name of the stage they are
    testing, eg. the basic collapse tests from :class:`BaseCollapseTestsMixin` will
    use the `collapse` dictionary from the subclass. The dictionary can be contain
    arbitrary values for configurating the tests.

    The basic mixins use `params` defined in the dict for each subcommand
    to pass arbitrary commandline arguments.

    When a key is not found in the dictionary, the test will look for a key
    in the test configuration yaml file. This is used in the mixin class for eg.
    `design` and `sample_ids` which are across multiple mixin classes.
    """

    # This will be injected by a fixture
    context: PixelatorWorkflowContext

    @property
    def workdir(self) -> Path:
        """Return the workdir from the workflow context."""
        return self.context.workdir



"""Functions for plotting component layouts for MPX data.

Copyright © 2024 Pixelgen Technologies AB.
"""

from typing import List, Literal, Tuple, Union

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import polars as pl
import scipy
from matplotlib import cm
from matplotlib.colors import Normalize

from pixelator.common.graph.backends.protocol import SupportedLayoutAlgorithm
from pixelator.common.marks import experimental
from pixelator.mpx.graph import Graph
from pixelator.mpx.pixeldataset import PixelDataset
from pixelator.mpx.plot.constants import Color


def _unit_sphere_surface(horizontal_resolution, vertical_resolution):
    horizontal_angles = np.linspace(0, 2 * np.pi, horizontal_resolution)
    vertical_angles = np.linspace(0, np.pi, vertical_resolution)

    X = np.outer(np.cos(horizontal_angles), np.sin(vertical_angles))
    Y = np.outer(np.sin(horizontal_angles), np.sin(vertical_angles))
    Z = np.outer(np.ones(np.size(horizontal_angles)), np.cos(vertical_angles))
    return X, Y, Z


def _calculate_distance_to_unit_sphere_zones(coordinates, unit_sphere_surface):
    X, Y, Z = unit_sphere_surface
    zones_on_sphere_surface = np.stack([X, Y, Z], axis=2).reshape(-1, 3)
    return scipy.spatial.distance.cdist(
        zones_on_sphere_surface, coordinates, "euclidean"
    )


def _calculate_densities(coordinates, distance_cutoff, unit_sphere_surface):
    dist = _calculate_distance_to_unit_sphere_zones(
        coordinates=coordinates, unit_sphere_surface=unit_sphere_surface
    )
    raw_sums = np.sum(
        1 - (dist / distance_cutoff), where=(dist < distance_cutoff), axis=1
    )
    raw_sums[raw_sums <= 1] = 0
    densities = np.log(raw_sums, out=raw_sums, where=(raw_sums > 1))
    densities = densities / np.max(densities)
    return densities


def _get_component_graph(pxl_data: PixelDataset, component: str):
    component_edges = pxl_data.edgelist_lazy.filter(pl.col("component") == component)
    component_graph = Graph.from_edgelist(
        component_edges,
        add_marker_counts=True,
        simplify=True,
        use_full_bipartite=True,
    )
    return component_graph


def _get_coordinates(
    pxl_data: PixelDataset,
    component: str,
    layout_algorithm: SupportedLayoutAlgorithm | None = None,
    cache_layout: bool = False,
    show_b_nodes: bool = False,
    random_seed: int | None = None,
) -> pd.DataFrame:
    component_graph = _get_component_graph(pxl_data=pxl_data, component=component)
    if (
        layout_algorithm is None
        and hasattr(pxl_data, "precomputed_layouts")
        and pxl_data.precomputed_layouts is not None
    ):
        coordinates = pxl_data.precomputed_layouts.filter(
            component_ids=component
        ).to_df()
    else:
        if layout_algorithm is None:
            layout_algorithm = "pmds_3d"
        coordinates = component_graph.layout_coordinates(
            layout_algorithm=layout_algorithm,  # type: ignore
            cache=cache_layout,
            only_keep_a_pixels=not show_b_nodes,
            random_seed=random_seed,
        )

    filtered_coordinates = coordinates
    filtered_coordinates["pixel_type"] = [
        component_graph.raw.nodes[ind]["pixel_type"]
        for ind in filtered_coordinates.index
    ]

    if not show_b_nodes:
        filtered_coordinates = filtered_coordinates[
            filtered_coordinates["pixel_type"] == "A"
        ]

    edgelist = pd.DataFrame(component_graph.es)
    edgelist = edgelist[edgelist[0].isin(filtered_coordinates.index)]
    edgelist = edgelist[edgelist[1].isin(filtered_coordinates.index)]
    edgelist = edgelist.loc[:, [0, 1]].to_numpy()

    return filtered_coordinates, edgelist, component_graph


def _plot_for_legend(coordinates: pd.DataFrame, axis, show_b_nodes, cmap, node_size):
    a_node_example = coordinates[coordinates["pixel_type"] == "A"].iloc[0, :]
    axis.scatter(
        a_node_example["x"],
        a_node_example["y"],
        c=0,
        cmap=cmap,
        vmin=0,
        s=node_size,
        label="A-nodes",
    )
    if show_b_nodes:
        b_node_example = coordinates[coordinates["pixel_type"] == "B"].iloc[0, :]
        axis.scatter(
            b_node_example["x"],
            b_node_example["y"],
            c=1,
            cmap=cmap,
            vmin=0,
            s=node_size,
            label="B-nodes",
        )


def _decorate_plot(
    im=None,
    include_colorbar: bool = False,
    vmax: float = 0,
    cmap: str = "cool",
    ax=None,
    fig=None,
    legend_ax=None,
):
    if include_colorbar:
        if isinstance(ax, np.ndarray):
            fig.colorbar(
                im,
                ax=ax.ravel().tolist(),
                cmap=cmap,
                norm=Normalize(vmin=0, vmax=vmax),
            )
        else:
            fig.colorbar(
                im,
                ax=ax,
                cmap=cmap,
                norm=Normalize(vmin=0, vmax=vmax),
            )
    if legend_ax is not None:
        if not include_colorbar:
            handles, labels = legend_ax.get_legend_handles_labels()
            fig.legend(handles, labels, loc="right")
            plt.subplots_adjust(right=0.80)
        else:
            raise AssertionError(
                "Plotting pixel_type together with other markers is not supported."
            )


def _get_current_axis(ax, i_m, i_c, marker_list, component):
    if len(marker_list) == 1:
        if len(component) == 1:
            crnt_ax = ax
        else:
            crnt_ax = ax[i_c]
    else:
        if len(component) == 1:
            crnt_ax = ax[i_m]
        else:
            crnt_ax = ax[i_m, i_c]

    if i_c == 0:  # Set the y-label only for the first column
        crnt_ax.set_ylabel(marker_list[i_m], rotation=90, size="large")
    if i_m == 0:  # Set the title only for the first row
        crnt_ax.set_title(component[i_c])

    return crnt_ax


def _get_color_values(mark, filtered_coordinates, log_scale, vmax):
    if mark is None:
        color_val = Color.NETWORKX_NODE_COLOR
    elif mark == "pixel_type":
        color_val = filtered_coordinates["pixel_type"] == "B"
    elif mark not in filtered_coordinates.columns:
        raise AssertionError(f"Marker {mark} not found in the component graph.")
    else:
        if log_scale:
            color_val = np.log1p(filtered_coordinates.loc[:, mark])
        else:
            color_val = filtered_coordinates.loc[:, mark]
        vmax = max(vmax, np.max(color_val))

    return color_val, vmax


def plot_2d_graph(
    pxl_data: PixelDataset,
    component: str | List[str],
    marker: str | List[str] = "pixel_type",
    layout_algorithm: SupportedLayoutAlgorithm | None = None,
    show_edges: bool = False,
    log_scale: bool = True,
    node_size: float = 10.0,
    edge_width: float = 1.0,
    show_b_nodes: bool = False,
    cmap: str = "cool",
    alpha: float = 0.5,
    cache_layout: bool = False,
    random_seed: int | None = None,
) -> Tuple[plt.Figure, plt.Axes | np.ndarray]:
    """Plot a (collection of) 2D graph(s) based on the given pixel data.

    The graph can be plotted for one or a list of components.
    The graph nodes can be colored by a marker. The marker can be a (list of) marker(s) or "pixel_type".
    Example usage: plot_2d_graph(pxl, component=["PXLCMP0000000"], marker=["HLA-ABC", "HLA-RA"]).

    :param pxl_data: The pixel dataset to plot.
    :param component: The component(s) to plot. Defaults to None.
    :param marker: The marker attribute to use for coloring the nodes. Defaults to "pixel_type".
    :param layout_algorithm: The layout algorithm to use. Defaults to None (checking for pre-computed).
    :param show_edges: Whether to show the edges in the graph. Defaults to False.
    :param log_scale: Whether to use a logarithmic scale for the marker attribute. Defaults to True.
    :param node_size: The size of the nodes. Defaults to 10.0.
    :param edge_width: The width of the edges. Defaults to 1.0.
    :param show_b_nodes: Whether to show the B-nodes. Defaults to False.
    :param cmap: The colormap to use for coloring the nodes. Defaults to "cool".
    :param alpha: The alpha value for the nodes. Defaults to 0.7.
    :param cache_layout: Whether to cache the layout coordinates. Defaults to False.
    :param random_seed: The random seed to use for the layout algorithm. Defaults to None.

    :return: The figure and axes objects of the plot.
    :rtype: Tuple[plt.Figure, plt.Axes]

    :raises: AssertionError if the marker is not found in the component graph.
    :raises: AssertionError if no nodes are found with the specified marker.
    :raises: AssertionError if "pixel_type" is in the markers together with other markers.

    """
    if isinstance(component, str):
        component = [component]

    if isinstance(marker, str):
        marker_list = [marker]
    else:
        marker_list = marker

    fig, ax = plt.subplots(nrows=len(marker_list), ncols=len(component))

    include_colorbar = False
    vmax = 0  # maximum value for colorbar
    for i_c, comp in enumerate(component):
        coordinates, edgelist, component_graph = _get_coordinates(
            pxl_data=pxl_data,
            component=comp,
            layout_algorithm=layout_algorithm,
            cache_layout=cache_layout,
            show_b_nodes=show_b_nodes,
            random_seed=random_seed,
        )

        for i_m, mark in enumerate(marker_list):
            crnt_ax = _get_current_axis(ax, i_m, i_c, marker_list, component)
            color_val, vmax = _get_color_values(mark, coordinates, log_scale, vmax)
            if mark != "pixel_type":
                include_colorbar = True

            im = nx.draw_networkx(
                component_graph.raw,
                nodelist=coordinates.index,
                pos=coordinates.loc[:, ["x", "y"]].T.to_dict("list"),
                ax=crnt_ax,
                node_size=node_size,
                node_color=color_val,
                cmap=cmap,
                width=edge_width,
                with_labels=False,
                edgelist=edgelist if show_edges else [],
                label="_nolegend_",
                alpha=alpha,
            )

            if mark == "pixel_type":
                # Re-plot one point from each pixel type to add a legend
                _plot_for_legend(coordinates, crnt_ax, show_b_nodes, cmap, node_size)
                legend_ax = crnt_ax
            else:
                legend_ax = None

            crnt_ax.grid(False)
            crnt_ax.spines[:].set_visible(False)
            crnt_ax.set_xticks([])
            crnt_ax.set_yticks([])

    _decorate_plot(
        im=im,
        include_colorbar=include_colorbar,
        vmax=vmax,
        cmap=cmap,
        ax=ax,
        fig=fig,
        legend_ax=legend_ax,
    )

    return fig, ax


def plot_3d_from_coordinates(
    coordinates: pd.DataFrame,
    node_size: float = 3.0,
    opacity: float = 0.4,
    cmap: str = "Inferno",
    suppress_fig: bool = False,
) -> go.Figure:
    """Plot a 3D graph from the given coordinates.

    :param coordinates: The coordinates to plot.
    :param node_size: The size of the nodes. Defaults to 3.0.
    :param opacity: The opacity of the nodes. Defaults to 0.4.
    :param cmap: The colormap to use for coloring the nodes. Defaults to "Inferno".
    :param suppress_fig: Whether to suppress (i.e. not plot) the figure. Defaults to False.
    :return: The plotted 3D graph.
    :rtype: go.Figure

    """
    fig = go.Figure(
        data=[
            go.Scatter3d(
                x=coordinates["x"],
                y=coordinates["y"],
                z=coordinates["z"],
                mode="markers",
                marker=dict(
                    size=node_size,
                    color=coordinates["color"],
                    colorscale=cmap,
                    opacity=opacity,
                    colorbar=dict(thickness=20, title="color"),
                ),
            )
        ]
    )

    if not suppress_fig:
        fig.show()
    return fig


def plot_3d_graph(
    pxl_data: PixelDataset,
    component: str,
    marker: Union[list, None] = None,
    layout_algorithm: SupportedLayoutAlgorithm | None = None,
    log_scale: bool = True,
    normalize: bool = False,
    node_size: float = 3.0,
    opacity: float = 0.4,
    show_b_nodes: bool = False,
    cmap: str = "Inferno",
    cache_layout: bool = False,
    suppress_fig: bool = False,
) -> go.Figure:
    """Plot a 3D graph of the specified component in the given PixelDataset.

    :param pxl_data: The PixelDataset containing the data.
    :param component: The component to plot.
    :param marker: The marker to use for coloring the nodes. Defaults to None.
    :param layout_algorithm: The layout algorithm to use for positioning the nodes. Defaults to "fruchterman_reingold_3d".
    :param log_scale: Whether to apply logarithmic scaling to the marker values. Defaults to True.
    :param normalize: Whether to normalize the coordinates. Defaults to False.
    :param node_size: The size of the nodes. Defaults to 3.0.
    :param opacity: The opacity of the nodes. Defaults to 0.4.
    :param show_b_nodes: Whether to show nodes of type B. Defaults to False.
    :param cmap: The colormap to use for coloring the nodes. Defaults to "Inferno".
    :param cache_layout: Whether to cache the layout coordinates. Defaults to False.
    :param suppress_fig: Whether to suppress (i.e. not plot) the figure. Defaults to False.
    :return: The plotted 3D graph.
    :rtype: go.Figure

    """
    coordinates, _, _ = _get_coordinates(
        pxl_data=pxl_data,
        component=component,
        layout_algorithm=layout_algorithm,
        cache_layout=cache_layout,
        show_b_nodes=show_b_nodes,
    )

    if marker is not None:
        if log_scale:
            coordinates["color"] = np.log1p(coordinates[marker])
        else:
            coordinates["color"] = coordinates[marker]
    else:
        coordinates["color"] = Color.SKYBLUE3

    if normalize:
        coordinates[["x", "y", "z"]] = coordinates[["x_norm", "y_norm", "z_norm"]]

    fig = plot_3d_from_coordinates(
        coordinates=coordinates,
        node_size=node_size,
        opacity=opacity,
        cmap=cmap,
        suppress_fig=True,
    )

    fig.update_layout(title=component)

    if not suppress_fig:
        fig.show()

    return fig


@experimental
def plot_3d_heatmap(
    component_graph: Graph,
    marker: str,
    distance_cutoff: float,
    layout_algorithm: Literal[
        "fruchterman_reingold_3d", "kamada_kawai_3d"
    ] = "fruchterman_reingold_3d",
    cache_layout: bool = False,
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot a 3D heatmap for the marker in the provided component.

    :param component_graph: A component graph to plot for.
    :param marker: marker to plot this for.
    :param distance_cutoff: a distance cutoff to use for determining size of
                            area to consider as close in the density calculation.
    :param layout_algorithm: Layout algorithm to use. Options are:
                            "fruchterman_reingold_3d" and "kamada_kawai_3d"
    :param cache_layout: set this to `True` to cache the layout
                         or faster computations on subsequent calls. This comes at the
                         cost of additional memory usage.
    :return: A matplotlib 3D heatmap figure, and it's associated Axes instance
    :rtype: Tuple[plt.Figure, plt.Axes]
    :raises: AssertionError if the provided `layout_algorithm` is not valid,
             or there are no with markers for the provided `marker`
    """
    coordinates = component_graph.layout_coordinates(
        layout_algorithm=layout_algorithm, cache=cache_layout
    )
    coordinates = coordinates[coordinates[marker] > 0]
    coordinates = np.array(coordinates[["x_norm", "y_norm", "z_norm"]])
    if len(coordinates) < 1:
        raise AssertionError(f"No nodes found with {marker}.")

    horizontal_resolution = 120
    vertical_resolution = 80
    X, Y, Z = _unit_sphere_surface(
        horizontal_resolution=horizontal_resolution,
        vertical_resolution=vertical_resolution,
    )

    densities = _calculate_densities(
        coordinates=coordinates,
        distance_cutoff=distance_cutoff,
        unit_sphere_surface=(X, Y, Z),
    )

    fig = plt.figure(figsize=(6, 6))
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    color_scale = cm.inferno(  # type: ignore
        densities.reshape(horizontal_resolution, -1)
    )
    ax.plot_surface(  # type: ignore
        X,
        Y,
        Z,
        cstride=1,
        rstride=1,
        facecolors=color_scale,
        shade=False,
    )
    ax.set_axis_off()
    ax.set_box_aspect([1.0, 1.0, 1.0])  # type: ignore
    return fig, ax



"""Analysis engine capable of running a list of analysis functions on each component in a pixeldataset.

Copyright © 2024 Pixelgen Technologies AB.
"""

import itertools
import logging
import multiprocessing
import typing
from collections import defaultdict
from dataclasses import dataclass
from functools import wraps
from logging.handlers import SocketHandler
from pathlib import Path
from typing import Callable, Generic, Iterable, Protocol, TypeVar

import click
import pandas as pd
import polars as pl
from joblib import Parallel, delayed

from pixelator.pna import read
from pixelator.pna.graph import PNAGraph
from pixelator.pna.pixeldataset import Component, PNAPixelDataset
from pixelator.pna.pixeldataset.io import PxlFile

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class LoggingSetup:
    """Dataclass to hold the logging setup for the analysis engine.

    Assumes that the logging setup uses a socket handler on local host
    to manage the logging.
    """

    port: int
    log_level: int

    @staticmethod
    def from_logger(logger: logging.Logger):
        """Create a LoggingSetup from a logger."""
        return LoggingSetup(port=logger.port, log_level=logger.log_level)  # type: ignore


def _add_handlers_to_root_logger(logging_setup):
    root_logger = logging.getLogger()

    if not root_logger.handlers:
        root_logger.setLevel(logging_setup.log_level)
        socket_handler = SocketHandler("localhost", logging_setup.port)
        root_logger.addHandler(socket_handler)


class _ParallelWithLogging(Parallel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _print(self, mgs):
        logger.debug(mgs)


def _get_joblib_executor(nbr_cores=None, **kwargs) -> Parallel:
    """Return a joblib executor with some default settings."""
    current_click_context = click.get_current_context(silent=True)
    click_nbr_cores = None
    if current_click_context:
        click_nbr_cores = current_click_context.obj.get("CORES")

    nbr_cores = (
        nbr_cores if nbr_cores else click_nbr_cores or multiprocessing.cpu_count()
    )
    return _ParallelWithLogging(n_jobs=nbr_cores, **kwargs)


T = TypeVar("T")


def with_logging(f):
    """Add logging to a function with this decorator.

    This is necessary to deal with the fact that for out of process workers
    like the ones used by joblib we need to set up logging in each worker.
    This decorator makes this slightly easier.
    """

    @wraps(f)
    def wrapper(*args, **kwds):
        # Turn this into a dectorator
        logging_setup = kwds.pop("logging_setup", None)
        if logging_setup:
            _add_handlers_to_root_logger(logging_setup)

        return f(*args, **kwds)

    return wrapper


class PerComponentTask(Protocol, Generic[T]):
    """Protocol for tasks that are run on each component in a PixelDataset.

    :var TASK_NAME: The name of the analysis.
    """

    TASK_NAME: typing.ClassVar[str]

    def set_dataset(self, pxl_file_path: Path):
        """Specify a dataset to enable analysis being run directly from component IDs."""
        ...

    @with_logging
    def run_from_component_id(self, component_id: str) -> T:
        """Run the analysis on this component.

        This method assumes that the dataset is stored internally
        and the component is directly accessible from its id.

        :param component_id: The id of the component.
        """
        raise NotImplementedError

    @with_logging
    def run_on_component_graph(self, component: PNAGraph, component_id: str) -> T:
        """Run the analysis on this component.

        :param component: The graph of the component.
        :param component_id: The id of the component.
        """
        raise NotImplementedError

    @with_logging
    def run_on_component_edgelist(
        self, component: pl.LazyFrame, component_id: str
    ) -> T:
        """Run the analysis on this component.

        :param component: The edgelist of the component.
        :param component_id: The id of the component.
        """
        raise NotImplementedError

    @with_logging
    def run_on_component(
        self, component: Component | str, logging_setup: LoggingSetup | None = None
    ) -> T:
        """Run the analysis on a component.

        This method will let the Analysis decide whether to run on a Graph or a LazyFrame.
        The default implementation will first try the graph based analysis and then fall back
        to the edgelist based analysis if the graph based analysis raises NotImplementedError.

        :param component: The component to run the analysis on. Either a Graph, a LazyFrame or the name of a component.
        :param logging_setup: The logging setup to use.
        :raises TypeError: If the component is not a Graph or a LazyFrame.
        """
        if isinstance(component, str):
            return self.run_from_component_id(component)
        try:
            return self.run_on_component_edgelist(
                component.frame, component.component_id
            )
        except NotImplementedError:
            pass
        try:
            return self.run_on_component_graph(component.graph, component.component_id)
        except NotImplementedError as e:
            logger.error(
                "Either `run_on_component_graph` or `run_on_component_edgelist` "
                "must be implemented by a `PerComponentTask`"
            )
            raise e

    def concatenate_data(self, data: Iterable[T]) -> T:
        """Concatenate the data. Override this if you need custom concatenation behavior."""
        try:
            scores = pd.concat(data, axis=0)
            return scores
        except ValueError as error:
            logger.error(f"No data was found to compute {self.TASK_NAME}")
            raise error

    def post_process_data(self, data: T) -> T:
        """Post process the data (e.g. adjust p-values). Override this if your data needs post processing."""
        return data

    def add_to_pixel_file(self, data: T, pxl_file_target: PxlFile) -> None:
        """Add the data in the right place in the pxl_dataset."""
        ...

    def parameters(self) -> dict[str, typing.Any]:
        """Return the parameters of the `PerComponentAnalysis`.

        This is used e.g. to store the metadata the parameters of the analysis
        in the run metadata.
        """
        return {f"{self.TASK_NAME}": vars(self)}


class AnalysisManager:
    """Analysis manager that can run a number of analysis across a stream of components.

    The analysis manager is responsible for hooking up the analysis functions and making
    them run on each component in the stream. The main workflow it uses is outlined in the
    `execute` method.
    """

    def __init__(
        self,
        analysis_to_run: Iterable[PerComponentTask],
        logging_setup: LoggingSetup | None = None,
        n_cores: int | None = None,
        pxl_dataset_builder: Callable[[Iterable[PxlFile]], PNAPixelDataset]
        | None = None,
    ):
        """Initialize the analysis manager.

        :param analysis_to_run: The analysis to run on each component.
        :param logging_setup: The logging setup to use.
        :param n_cores: The number of cores to use for parallel processing (set to 1 to disable parallel processing).
        :param pxl_dataset_builder: A function that can build a PixelDataset from an iterable of PxlFiles.
        """
        self.analysis_to_run = {
            analysis.TASK_NAME: analysis for analysis in analysis_to_run
        }
        self._logging_setup = logging_setup
        if n_cores is not None and n_cores < 1:
            raise ValueError("n_cores must be greater than 0 or None.")
        self._n_cores = n_cores
        if pxl_dataset_builder is None:
            pxl_dataset_builder = PNAPixelDataset.from_pxl_files
        self._pxl_dataset_builder = pxl_dataset_builder

    def _execute_computations_in_parallel(
        self, component_stream: Iterable[Component | str]
    ):
        with _get_joblib_executor(
            nbr_cores=self._n_cores, verbose=100, return_as="generator_unordered"
        ) as parallel:

            def func(component, analysis_to_run):
                results = []
                for analysis_name, analysis in analysis_to_run.items():
                    logger.debug(
                        f"running {analysis_name} on {component.component_id if isinstance(component, Component) else component}"
                    )
                    results.append(
                        (
                            analysis_name,
                            analysis.run_on_component(
                                component, logging_setup=self._logging_setup
                            ),
                        )
                    )

                return results

            yield from itertools.chain.from_iterable(
                parallel(
                    delayed(func)(
                        component,
                        self.analysis_to_run,
                    )
                    for component in component_stream
                )
            )

    def _execute_computations_sequentially(
        self, component_stream: Iterable[Component | str]
    ):
        for component in component_stream:
            results = []
            for analysis_name, analysis in self.analysis_to_run.items():
                logger.debug(
                    f"running {analysis_name} on {component.component_id if isinstance(component, Component) else component}"
                )
                results.append(
                    (
                        analysis_name,
                        analysis.run_on_component(
                            component, logging_setup=self._logging_setup
                        ),
                    )
                )
            yield from results

    def _post_process(self, per_component_results):
        concatenated_data = defaultdict(list)
        for key, data in per_component_results:
            concatenated_data[key].append(data)

        for key, data_list in concatenated_data.items():
            yield (
                key,
                self.analysis_to_run[key].post_process_data(
                    self.analysis_to_run[key].concatenate_data(data_list)
                ),
            )

    def _add_to_pixel_dataset(
        self, post_processed_data, pxl_file_target: PxlFile
    ) -> PNAPixelDataset:
        for key, data in post_processed_data:
            self.analysis_to_run[key].add_to_pixel_file(data, pxl_file_target)
        return self._pxl_dataset_builder(pxl_file_target)  # type: ignore

    def _execute_on_iterator(self, iterator, pxl_file_target: PxlFile):
        if self._n_cores is None or self._n_cores > 1:
            per_component_results = self._execute_computations_in_parallel(iterator)
        else:
            per_component_results = self._execute_computations_sequentially(iterator)
        post_processed_data = self._post_process(per_component_results)
        pxl_dataset_with_results = self._add_to_pixel_dataset(
            post_processed_data, pxl_file_target=pxl_file_target
        )
        return pxl_dataset_with_results

    def execute(
        self, pixel_dataset: PNAPixelDataset, pxl_file_target: PxlFile
    ) -> PNAPixelDataset:
        """Execute the analysis on the provided pixel dataset."""
        iterator = pixel_dataset.edgelist().iterator()
        return self._execute_on_iterator(iterator, pxl_file_target)

    def _set_path_to_dataset(self, input_pxl_file_path: Path):
        for key in self.analysis_to_run.keys():
            self.analysis_to_run[key].set_dataset(input_pxl_file_path)

    def execute_from_path(
        self, input_pxl_file_path: Path, pxl_file_target: PxlFile
    ) -> PNAPixelDataset:
        """Execute the analysis on the provided pixel file path."""
        self._set_path_to_dataset(input_pxl_file_path)
        pixel_dataset = read(input_pxl_file_path)
        iterator = pixel_dataset.components()
        return self._execute_on_iterator(iterator, pxl_file_target)



"""Copyright © 2023 Pixelgen Technologies AB."""

import logging

import pytest

from .base import BaseWorkflowTestMixin

logger = logging.getLogger(__name__)


class BaseAnalysisTestsMixin(BaseWorkflowTestMixin):
    """Base class for analysis command tests.

    Test cases (defined in this class or in subclasses)
    that depend on the output should be marked with:
    ```
    @pytest.mark.dependency(depends=["test_analysis_run"])
    ```
    """

    __stage_key__ = "analysis"

    @pytest.mark.dependency(name="test_analysis_run", depends=["test_annotate_run"])
    def test_analysis_run(self):
        params = self.__get_parameters()
        verbose = self.__get_options("common").get("verbose")
        input_files = list(
            map(
                lambda f: str(f),
                (self.workdir / "annotate").glob("*.annotate.dataset.pxl"),
            )
        )

        command = [
            "--log-file",
            "analysis-pixelator.log",
            "single-cell-mpx",
            "analysis",
            "--output",
            ".",
        ]

        if verbose:
            command = ["--verbose"] + command

        if params:
            command += params

        command += input_files

        self.context.run_command("analysis", command, input_files)
        assert self.__this_result.exit_code == 0

    @pytest.mark.dependency(depends=["test_analysis_run"])
    def test_analysis_dataset_exists(self):
        pxl_files = (self.workdir / "analysis").glob("*.analysis.dataset.pxl")
        for f in pxl_files:
            assert f.is_file()

    @pytest.mark.dependency(depends=["test_analysis_run"])
    def test_analysis_report_exists(self):
        json_files = (self.workdir / "analysis").glob("*.report.json")
        for f in json_files:
            assert f.is_file()



"""Collect files and transform data for reporting from a :py:class:`PixelatorWorkdir`.

Copyright © 2023 Pixelgen Technologies AB.
"""

from __future__ import annotations

import logging
import typing
from pathlib import Path
from typing import Iterable

import pandas as pd

from pixelator.mpx.report.common.cli_info import CLIInvocationInfo
from pixelator.mpx.report.common.workdir import (
    PixelatorWorkdir,
    SingleCellStage,
    WorkdirOutputNotFound,
)
from pixelator.mpx.report.models import (
    AdapterQCSampleReport,
    AmpliconSampleReport,
    AnalysisSampleReport,
    AnnotateSampleReport,
    CollapseSampleReport,
    CommandInfo,
    DemuxSampleReport,
    GraphSampleReport,
    MoleculesDataflowReport,
    PreQCSampleReport,
    ReadsDataflowReport,
)
from pixelator.mpx.report.models.base import SampleReport
from pixelator.mpx.report.models.layout import LayoutSampleReport

logger = logging.getLogger("pixelator.report")
ModelT = typing.TypeVar("ModelT", bound=type[SampleReport])


class PixelatorReporting:
    """Collect files for reporting from a :py:class:`PixelatorWorkdir`.

    :ivar workdir: The pixelator output folder or a :py:class:`PixelatorWorkdir`
        instance
    """

    def __init__(self, workdir: Path | PixelatorWorkdir):
        """Initialize the PixelatorReporting object.

        :param workdir: The pixelator output folder or a :py:class:`PixelatorWorkdir`
        """
        if isinstance(workdir, PixelatorWorkdir):
            self.workdir = workdir
        else:
            self.workdir = PixelatorWorkdir(workdir)

        # all pixelator commands in defined order
        self._command_list: list[str] | None = None

    def samples(self) -> set[str]:
        """Return a list of all samples encountered from command metadata files."""
        return self.workdir.samples()

    @staticmethod
    def _combine_data(reports, model: ModelT) -> pd.DataFrame:
        """Merge a list of :class:`Model` objects into a single dataframe."""
        data = []

        for r in reports:
            logger.debug("Parsing metrics file: %s", r)
            data.append(model.from_json(r).model_dump())

        df = pd.DataFrame(data)
        df.astype({"sample_id": "string"})
        df.set_index("sample_id", inplace=True)
        df.sort_index(inplace=True)
        return df

    @staticmethod
    def _explode_json_columns(
        df: pd.DataFrame, keys_to_explode: Iterable[str]
    ) -> pd.DataFrame:
        """Explode json columns of a dataframe.

        :param df: The dataframe to explode
        :param keys_to_explode: The keys of the json columns to explode
        """
        series_to_concat = []
        for key in keys_to_explode:
            dict_column = pd.json_normalize(df[key])
            cleaned_key = key.replace("_stats", "")
            dict_column.columns = [
                f"{cleaned_key}_{col}" for col in dict_column.columns
            ]
            dict_column.set_index(df.index, inplace=True)
            series_to_concat.append(dict_column)

        # Merge the exploded dict into the original dataframe and drop the dict column
        df.drop(
            labels=list(keys_to_explode),
            axis="columns",
            inplace=True,
        )
        df = pd.concat(
            (df, *series_to_concat),
            join="inner",
            axis="columns",
        )
        return df

    def amplicon_metrics(self, sample_name: str) -> AmpliconSampleReport:
        """Return the amplicon metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.AMPLICON, sample_name
        )
        return AmpliconSampleReport.from_json(sample_file)

    def amplicon_summary(self) -> pd.DataFrame:
        """Combine all amplicon reports into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.AMPLICON)
        df = self._combine_data(reports, AmpliconSampleReport)
        return df

    def preqc_metrics(self, sample_name: str) -> PreQCSampleReport:
        """Return the preqc metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.PREQC, sample_name
        )
        return PreQCSampleReport.from_json(sample_file)

    def preqc_summary(self) -> pd.DataFrame:
        """Combine all preqc reports into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.PREQC)
        df = self._combine_data(reports, PreQCSampleReport)
        return df

    def adapterqc_metrics(self, sample_name: str) -> AdapterQCSampleReport:
        """Return the adapterqc metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.ADAPTERQC, sample_name
        )
        return AdapterQCSampleReport.from_json(sample_file)

    def adapterqc_summary(self) -> pd.DataFrame:
        """Combine all preqc reports into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.ADAPTERQC)
        df = self._combine_data(reports, AdapterQCSampleReport)
        return df

    def demux_metrics(self, sample_name: str) -> DemuxSampleReport:
        """Return the demux metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.DEMUX, sample_name
        )
        return DemuxSampleReport.from_json(sample_file)

    def demux_summary(self) -> pd.DataFrame:
        """Combine all demux reports into a single dataframe.

        For example:
                            input_read_count  output_read_count ACTB ...  mIgG2b
        sample_id                                                   55 ...      89
        pbmcs_unstimulated             199390              189009   86 ...     115
        """
        reports = self.workdir.single_cell_report(SingleCellStage.DEMUX)
        df = self._combine_data(reports, DemuxSampleReport)

        # We do not use _explode_json_columns here because the
        # keys naming is different here

        # Extract the dict column from the dataframe and expand each key
        # into a new column
        read_dict = pd.json_normalize(df["per_antibody_read_counts"])
        read_dict.columns = [f"read_counts_{col}" for col in read_dict.columns]
        read_dict.set_index(df.index, inplace=True)

        read_fraction_dict = pd.json_normalize(df["per_antibody_read_count_fractions"])
        read_fraction_dict.columns = [
            f"read_count_fraction_{col}" for col in read_fraction_dict.columns
        ]
        read_fraction_dict.set_index(df.index, inplace=True)

        # Merge the exploded dict into the original dataframe and drop the dict column
        df.drop(
            labels=["per_antibody_read_counts", "per_antibody_read_count_fractions"],
            axis="columns",
            inplace=True,
        )
        df = pd.concat(
            (df, read_dict, read_fraction_dict), join="inner", axis="columns"
        )
        return df

    def collapse_metrics(self, sample_name: str) -> CollapseSampleReport:
        """Return the collapse metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.COLLAPSE, sample_name
        )
        return CollapseSampleReport.from_json(sample_file)

    def collapse_summary(self) -> pd.DataFrame:
        """Combine all collapse sumaries into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.COLLAPSE)
        df = self._combine_data(reports, CollapseSampleReport)

        keys_to_explode = ["collapsed_molecule_count_stats"]
        df = self._explode_json_columns(df, keys_to_explode)
        return df

    def graph_metrics(self, sample_name: str) -> GraphSampleReport:
        """Return the graph metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.GRAPH, sample_name
        )
        return GraphSampleReport.from_json(sample_file)

    def graph_summary(self) -> pd.DataFrame:
        """Combine graph metrics for all samples into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.GRAPH)
        df = self._combine_data(reports, GraphSampleReport)

        # Extract the dict column from the dataframe and expand each key
        # into a new column
        keys_to_explode = [
            "read_count_per_molecule_stats",
        ]

        df = self._explode_json_columns(df, keys_to_explode)
        return df

    def annotate_metrics(self, sample_name: str) -> AnnotateSampleReport:
        """Return the annotate metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.ANNOTATE, sample_name
        )
        return AnnotateSampleReport.from_json(sample_file)

    def annotate_summary(self) -> pd.DataFrame:
        """Combine graph metrics for all samples into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.ANNOTATE)
        df = self._combine_data(reports, AnnotateSampleReport)

        keys_to_explode = [
            "molecule_count_per_cell_stats",
            "read_count_per_cell_stats",
            "a_pixel_count_per_cell_stats",
            "b_pixel_count_per_cell_stats",
            "marker_count_per_cell_stats",
            "a_pixel_b_pixel_ratio_per_cell_stats",
            "molecule_count_per_a_pixel_per_cell_stats",
            "b_pixel_count_per_a_pixel_per_cell_stats",
            "a_pixel_count_per_b_pixel_per_cell_stats",
        ]

        df = self._explode_json_columns(df, keys_to_explode)
        return df

    def layout_metrics(self, sample_name: str) -> LayoutSampleReport:
        """Return the layout metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.LAYOUT, sample_name
        )
        return LayoutSampleReport.from_json(sample_file)

    def layout_summary(self) -> pd.DataFrame:
        """Combine graph metrics for all samples into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.LAYOUT)
        df = self._combine_data(reports, LayoutSampleReport)
        return df

    def analysis_metrics(self, sample_name: str) -> AnalysisSampleReport:
        """Return the analysis metrics for a sample."""
        sample_file = self.workdir.single_cell_report(
            SingleCellStage.ANALYSIS, sample_name
        )
        return AnalysisSampleReport.from_json(sample_file)

    def analysis_summary(self) -> pd.DataFrame:
        """Combine graph metrics for all samples into a single dataframe."""
        reports = self.workdir.single_cell_report(SingleCellStage.ANALYSIS)
        df = self._combine_data(reports, AnalysisSampleReport)
        keys_to_explode = ["polarization", "colocalization"]
        df = self._explode_json_columns(df, keys_to_explode)
        return df

    def reads_flow(self, sample_name: str) -> ReadsDataflowReport:
        """Return a summary with the flow of read counts through the pipeline.

        :param sample_name: The sample to return the dataflow for
        :return ReadsDataflowReport: A class:`ReadsDataflowReport` instance
        """
        preqc_metrics = self.preqc_metrics(sample_name)
        adapterqc_metrics = self.adapterqc_metrics(sample_name)
        demux_metrics = self.demux_metrics(sample_name)
        collapse_metrics = self.collapse_metrics(sample_name)
        annotate_metrics = self.annotate_metrics(sample_name)

        return ReadsDataflowReport(
            sample_id=sample_name,
            input_read_count=preqc_metrics.total_read_count,
            qc_filtered_read_count=preqc_metrics.passed_filter_read_count,
            valid_pbs_read_count=adapterqc_metrics.passed_filter_read_count,
            valid_antibody_read_count=demux_metrics.output_read_count,
            raw_molecule_read_count=collapse_metrics.output_read_count,
            size_filter_fail_molecule_read_count=annotate_metrics.size_filter_fail_read_count,
            aggregate_molecule_read_count=annotate_metrics.reads_in_aggregates_count,
            cell_molecule_read_count=annotate_metrics.read_count,
        )

    def reads_flow_summary(self):
        """Combine reads flow metrics for all samples into a single dataframe."""
        data = [self.reads_flow(sample).model_dump() for sample in self.samples()]

        df = pd.DataFrame(data)
        df.astype({"sample_id": "string"})
        df.set_index("sample_id", inplace=True)
        df.sort_index(inplace=True)

        return df

    def molecules_flow(self, sample_name: str) -> MoleculesDataflowReport:
        """Return a summary with the flow of molecules counts through the pipeline.

        :param sample_name: The sample to return the dataflow for
        :return MoleculesDataflowReport:
            A class:`MoleculesDataflowReport` instance
        """
        collapse_metrics = self.collapse_metrics(sample_name)
        annotate_metrics = self.annotate_metrics(sample_name)

        return MoleculesDataflowReport(
            sample_id=sample_name,
            raw_molecule_count=collapse_metrics.molecule_count,
            size_filter_fail_molecule_count=annotate_metrics.size_filter_fail_molecule_count,
            aggregate_molecule_count=annotate_metrics.molecules_in_aggregates_count,
            cell_molecule_count=annotate_metrics.molecule_count,
        )

    def molecules_flow_summary(self):
        """Combine molecule flow metrics for all samples into a single dataframe."""
        data = [self.molecules_flow(sample).model_dump() for sample in self.samples()]

        df = pd.DataFrame(data)
        df.astype({"sample_id": "string"})
        df.set_index("sample_id", inplace=True)
        df.sort_index(inplace=True)
        return df

    def cli_invocation_info(self, sample: str, cache: bool = True) -> CLIInvocationInfo:
        """Return the commandline options used to invoke multiple pixelator commands on a sample.

        :param sample: The sample to return the commandline for
        :param cache: Use cached data if available
        :return: A `CLIInvocationInfo` to query commandline invocations.
        :raises WorkdirOutputNotFound: If no commandline metadata is found for the sample
        """
        # Function scope import to avoid circular dependencies

        try:
            metadata_files = self.workdir.metadata_files(sample, cache=cache)
        except WorkdirOutputNotFound as e:
            e.message = f'No command line metadata found for sample: "{sample}"'
            raise

        data_flat = []
        for f in metadata_files:
            command_info_flat = CommandInfo.from_json(f)
            data_flat.append(command_info_flat)

        if self._command_list is None:
            self._command_list = _ordered_pixelator_commands()

        # mypy cannot detect that self._command_list cannot be None
        # anymore at this point so we cast it explicitly.
        data_flat.sort(
            key=lambda x: typing.cast("list[str]", self._command_list).index(x.command)
        )
        return CLIInvocationInfo(data_flat, sample_id=sample)


def _ordered_pixelator_commands() -> list[str]:
    """Return a list of pixelator CLI commands in depth-first order."""
    # local imports to avoid circular dependency issues
    import click

    from pixelator.cli import main_cli as click_context

    def build_command_list(obj: typing.Any, prefix=None):
        """Recursively go through groups and commands to create a list of commands."""
        prefix = prefix or []

        if isinstance(obj, click.Group):
            res = []
            new_prefix = prefix + [obj.name]

            for subcommand in obj.commands.values():
                res.extend(build_command_list(subcommand, prefix=new_prefix))

            return res

        return (" ".join(prefix + [obj.name]),)

    return build_command_list(click_context)
